"""
Python function text generator.
"""

try:
    # if running from a parent package (python -m app.py ...) i.e. relative import
    from .formats import (
        FORMAT_PY_FUNC_CREATE_TABLE,
        FORMAT_PY_FUNC_DELETE_FROM,
        FORMAT_PY_FUNC_INSERT_UPDATE,
        FORMAT_PY_FUNC_INSERT_UPDATE_FIELD,
        FORMAT_PY_FUNC_RENAME_COLUMN,
        TAB_SPACE_COUNT,
    )
    from .options import STATEMENT_TYPE_ALT_NAMES
except ImportError:
    # running from app folder (python app.py ...)
    from formats import (
        FORMAT_PY_FUNC_CREATE_TABLE,
        FORMAT_PY_FUNC_DELETE_FROM,
        FORMAT_PY_FUNC_INSERT_UPDATE,
        FORMAT_PY_FUNC_INSERT_UPDATE_FIELD,
        FORMAT_PY_FUNC_RENAME_COLUMN,
        TAB_SPACE_COUNT,
    )
    from options import STATEMENT_TYPE_ALT_NAMES


def gen_py_printable_statement(statement_type, table, lines, prefix=None):
    return f"{(prefix+'_') if prefix else ''}{STATEMENT_TYPE_ALT_NAMES[statement_type]}_{table.upper()} = \"\\\n{' ' * TAB_SPACE_COUNT}" +\
        f"\\\n{' ' * TAB_SPACE_COUNT}".join(lines) + "\\\n\""


def gen_py_db_create_table_function(table, prefix=None):
    return FORMAT_PY_FUNC_CREATE_TABLE % {
        "tableLower": table.lower(),
        "prefix_": f"{(prefix+'_') if prefix else ''}",
        "tableUpper": table.upper(),
    }


def gen_py_db_insert_update_function(table, field_names, count_fields, max_len_field_name, prefix=None):
    field_lines = ''
    extra_char_count = 3
    max_len_field_name += extra_char_count
    for inx, name in enumerate(field_names):
        sep0 = ' ' * (max_len_field_name + (TAB_SPACE_COUNT - max_len_field_name % TAB_SPACE_COUNT) - len(name) - extra_char_count)
        field_lines += FORMAT_PY_FUNC_INSERT_UPDATE_FIELD % {
            "colName": name,
            "sep0": sep0,
            "inx": inx,
            "comma": '' if inx==(count_fields-1) else ','
        }
    return (FORMAT_PY_FUNC_INSERT_UPDATE % {
        "tableLower": table.lower(),
        "prefix_": f"{(prefix+'_') if prefix else ''}",
        "tableUpper": table.upper(),
        "fieldLines": field_lines,
    }).replace("+>?X$;~;^&$", "%")


def gen_py_db_delete_function(table, column, prefix=None):
    return (FORMAT_PY_FUNC_DELETE_FROM % {
        "tableLower": table.lower(),
        "prefix_": f"{(prefix+'_') if prefix else ''}",
        "tableUpper": table.upper(),
        "column": column if column else 'id'
    }).replace("+>?X$;~;^&$", "%")


def gen_py_db_rename_column_function(table, prefix=None):
    return (FORMAT_PY_FUNC_RENAME_COLUMN % {
        "tableLower": table.lower(),
        "prefix_": f"{(prefix+'_') if prefix else ''}",
        "tableUpper": table.upper(),
    }).replace("+>?X$;~;^&$", "%")



"""
App (CLI) to generate MS SQL statement

Statement types supported:
- CREATE TABLE
- INSERT INTO
- MERGE
- DELETE FROM
- RENAME COLUMN

Arguments description:
- type:                 MS SQL Statement type
- table_descriptor:     json file with info describing the table (see example_td.json)
- column_name:          (optional) column to operate on
- where:                (optional) WHERE condition text

Example commands:
- (sql mode)     python app.py -t CREATE_TABLE -i example_td.json -o out/test.sql
- (python mode)  python app.py -t CREATE_TABLE -i example_td.json -o out/test.py -py -pre SQL
- (show help)    python app.py -h

More example commands:
- python app.py -t CREATE_TABLE -i example_td.json -o out/test.py -py -pre SQL -func
- python app.py -t INSERT_INTO -i example_td.json -o out/test.py -py -pre SQL -func
- python app.py -t MERGE -i example_td.json -o out/test.py -py -pre SQL -func
- python app.py -t DELETE_FROM -i example_td.json -o out/test.py -py -pre SQL -func -where "id = %(id)s" -col id
- python app.py -t RENAME_COLUMN -i example_td.json -o out/test.py -py -pre SQL -func

Can run from either:
- some parent package using "python -m <pkg1>.<pkg2>.<...>.app ..."
- app folder using "python app.py ..."
"""

try:
    # if running from a parent package (python -m app.py ...) i.e. relative import
    from .log import (
        sqllogger, 
        print_to_terminal_OFF, print_to_terminal_ON, 
        write_to_log_OFF, write_to_log_ON
    )
    from .options import (
        OPTIONS_STATEMENT_TYPE,
        OPTIONS_STATEMENT_TYPE_LOWER,
    )
    from .gen import (
        get_table_descriptor_from_file,
        ms_sql_statement,
    )
except ImportError:
    # running from app folder (python app.py ...)
    from log import (
        sqllogger, 
        print_to_terminal_OFF, print_to_terminal_ON, 
        write_to_log_OFF, write_to_log_ON
    )
    from options import (
        OPTIONS_STATEMENT_TYPE,
        OPTIONS_STATEMENT_TYPE_LOWER,
    )
    from gen import (
        get_table_descriptor_from_file,
        ms_sql_statement,
    )


if __name__=='__main__':
    write_to_log_OFF()
    print_to_terminal_ON()
    
    # define cli arguments
    import argparse
    parser = argparse.ArgumentParser(description="MS SQL Statement text generator.")
    parser.add_argument('--type', '-type', '-t', required=True, help='MS SQL statement type',\
        choices=OPTIONS_STATEMENT_TYPE + OPTIONS_STATEMENT_TYPE_LOWER)
    parser.add_argument('--table', '-table', '-td', '-i', required=True, help='table descriptor file (json)')
    parser.add_argument('--output', '-output', '-o', default=None, help='output file name')
    parser.add_argument('--overwrite', '-overwrite', '-ow', action='store_true', default=False, help='output file write mode')
    parser.add_argument('--python', '-python', '-py', action='store_true', default=False, help='enable python mode for output file')
    parser.add_argument('--prefix', '-prefix', '-pre', default=None, help='variable name prefix for python mode')
    parser.add_argument('--pyfunc', '-pyfunc', '-func', action='store_true', default=False,\
        help='generate python function alongside printed statement for executing printed statement')
    parser.add_argument('--column', '--col', '-column', '-col', '-c', default=None,\
        help='column (name) to operate on')
    parser.add_argument('--where', '-where', '-w', default=None,\
        help='WHERE condition text to append to final statement (or oldColumnName in case of RENAME COLUMN statement)')
    
    # parse arguments
    args = parser.parse_args()
    arg_type = args.type.upper()        # required
    arg_td_file = args.table            # required
    arg_out_file = args.output          # optional
    arg_overwrite = args.overwrite      # optional
    arg_py = args.python                # optional
    arg_py_pre = args.prefix            # optional
    arg_py_func = args.pyfunc           # optional
    arg_column = args.column            # optional
    arg_where = args.where              # optional 

    # read table descriptor file (json)
    success, message, td = get_table_descriptor_from_file(arg_td_file)
    if not success:
        sqllogger(message)
        exit()

    # call statement generator
    statement, message, print_statement = ms_sql_statement(
        arg_type,
        table_descriptor=td,
        column_name=arg_column,
        where=arg_where,
        out_file=arg_out_file,
        out_overwrite=arg_overwrite,
        printable=True,
        py_mode=arg_py,
        py_prefix=arg_py_pre,
        py_func=arg_py_func,
    )

    if message:
        sqllogger(message)

    # print statement to terminal
    sqllogger(print_statement)
    print_to_terminal_OFF()
    write_to_log_ON()



"""
Table-Descriptor-parsing and Statement-generating functions.
"""

try:
    # if running from a parent package (python -m app.py ...) i.e. relative import
    from .options import (
        CONSTRAINT_PRIMARY_KEY,
        STATEMENT_TYPE_CREATE_TABLE, 
        STATEMENT_TYPE_INSERT_INTO,
        STATEMENT_TYPE_MERGE,
        STATEMENT_TYPE_DELETE_FROM,
        STATEMENT_TYPE_RENAME_COLUMN,
        COL_MERGE_TYPE_ACCUMLATE,
        FILE_WRITE_MODE_APPEND,
        FILE_WRITE_MODE_OVERWITE,
        OPTIONS_STATEMENT_TYPE,
    )
    from .formats import (
        FORMAT_RENAME_COLUMN,
        proper_table_name,
        FORMAT_DELETE_FROM, FORMAT_DELETE_FROM_WHERE, FORMAT_INSERT_INTO_LINE_COL, FORMAT_INSERT_INTO_LINE_END, 
        FORMAT_INSERT_INTO_LINE_START, FORMAT_MERGE_LINE_COL, FORMAT_MERGE_LINE_END, FORMAT_MERGE_LINE_MATCH, 
        FORMAT_MERGE_LINE_MATCH_ACCUMUL, FORMAT_MERGE_LINE_MATCH_REPLACE, FORMAT_MERGE_LINE_NMATCH, 
        FORMAT_MERGE_LINE_NMATCH_FIELD, FORMAT_MERGE_LINE_NMATCH_VAL, FORMAT_MERGE_LINE_NMATCH_VALUES, 
        FORMAT_MERGE_LINE_ON, FORMAT_MERGE_LINE_ON_END, FORMAT_MERGE_LINE_ON_KEY, FORMAT_MERGE_LINE_START, 
        FORMAT_MERGE_LINE_USING, TAB_SPACE_COUNT, FORMAT_CREATE_TABLE_LINE_COL, FORMAT_CREATE_TABLE_LINE_END, 
        FORMAT_CREATE_TABLE_LINE_FK, FORMAT_CREATE_TABLE_LINE_FK_CONSTR, FORMAT_CREATE_TABLE_LINE_PKS, 
        FORMAT_CREATE_TABLE_LINE_START, FORMAT_CREATE_TABLE_LINE_TRUNC, FORMAT_DROP_TABLE_IF_EXISTS, 
    )
    from .tdparse import (
        parse_table_descriptor_all, 
        parse_table_descriptor_table_name, 
        parse_table_descriptor_table_name_fields
    )
    from .pyfunc import (
        gen_py_db_rename_column_function,
        gen_py_printable_statement,
        gen_py_db_create_table_function,
        gen_py_db_insert_update_function,
        gen_py_db_delete_function,
    )
except ImportError:
    # running from app folder (python app.py ...)
    from options import (
        CONSTRAINT_PRIMARY_KEY,
        STATEMENT_TYPE_CREATE_TABLE, 
        STATEMENT_TYPE_INSERT_INTO,
        STATEMENT_TYPE_MERGE,
        STATEMENT_TYPE_DELETE_FROM,
        STATEMENT_TYPE_RENAME_COLUMN,
        COL_MERGE_TYPE_ACCUMLATE,
        FILE_WRITE_MODE_APPEND,
        FILE_WRITE_MODE_OVERWITE,
        OPTIONS_STATEMENT_TYPE,
    )
    from formats import (
        FORMAT_RENAME_COLUMN,
        proper_table_name,
        FORMAT_DELETE_FROM, FORMAT_DELETE_FROM_WHERE, FORMAT_INSERT_INTO_LINE_COL, FORMAT_INSERT_INTO_LINE_END, 
        FORMAT_INSERT_INTO_LINE_START, FORMAT_MERGE_LINE_COL, FORMAT_MERGE_LINE_END, FORMAT_MERGE_LINE_MATCH, 
        FORMAT_MERGE_LINE_MATCH_ACCUMUL, FORMAT_MERGE_LINE_MATCH_REPLACE, FORMAT_MERGE_LINE_NMATCH, 
        FORMAT_MERGE_LINE_NMATCH_FIELD, FORMAT_MERGE_LINE_NMATCH_VAL, FORMAT_MERGE_LINE_NMATCH_VALUES, 
        FORMAT_MERGE_LINE_ON, FORMAT_MERGE_LINE_ON_END, FORMAT_MERGE_LINE_ON_KEY, FORMAT_MERGE_LINE_START, 
        FORMAT_MERGE_LINE_USING, TAB_SPACE_COUNT, FORMAT_CREATE_TABLE_LINE_COL, FORMAT_CREATE_TABLE_LINE_END, 
        FORMAT_CREATE_TABLE_LINE_FK, FORMAT_CREATE_TABLE_LINE_FK_CONSTR, FORMAT_CREATE_TABLE_LINE_PKS, 
        FORMAT_CREATE_TABLE_LINE_START, FORMAT_CREATE_TABLE_LINE_TRUNC, FORMAT_DROP_TABLE_IF_EXISTS, 
    )
    from tdparse import (
        parse_table_descriptor_all, 
        parse_table_descriptor_table_name, 
        parse_table_descriptor_table_name_fields
    )
    from pyfunc import (
        gen_py_db_rename_column_function,
        gen_py_printable_statement,
        gen_py_db_create_table_function,
        gen_py_db_insert_update_function,
        gen_py_db_delete_function,
    )
 

def gen_printable_statement(lines):
    return '\n'.join(lines)


def generate_statement_create_table(
        table_descriptor=None,
        column_name=None,
        where=None,
        printable=False,
        py_mode=False,
        py_prefix=None,
        py_func=None,
    ):

    statement = ''
    print_statement = ''

    # process table descriptor
    success, message, data = parse_table_descriptor_all(table_descriptor)

    if success:

        table_name, dependent_tables, field_names, field_types, field_constraints,\
            field_merge_types, relation_fields, relation_sources, relation_constraints,\
                count_fields, count_keys, max_len_field_name, max_len_field_type, count_fks,\
                    key_fields, non_key_fields, merge_on_fields, non_merge_on_fields,\
                        non_merge_on_fields_merge_types = data

        lines = []
        printable_lines = []

        def add_line(line, printable_line=None):
            lines.append(line)
            if printable: printable_lines.append(printable_line if printable_line else line)

        def gen_col_line(sep0, name, sep1, field_type, sep2, constraints, inx, count_fields):
            return FORMAT_CREATE_TABLE_LINE_COL % {
                "sep0": sep0,
                "colName": name,
                "sep1": sep1,
                "colType": field_type,
                "sep2": sep2 if constraints else '',
                "constraints": constraints,
                "comma": '' if inx==(count_fields-1) else ','
            }

        if dependent_tables:
            # dependent deletion lines
            for dep in dependent_tables:
                add_line(FORMAT_DROP_TABLE_IF_EXISTS % {'tableName': proper_table_name(dep)})
        # self deletion line
        add_line(FORMAT_DROP_TABLE_IF_EXISTS % {'tableName': proper_table_name(table_name)})
        # statement start line
        add_line(FORMAT_CREATE_TABLE_LINE_START % {'tableName': table_name})
        # field definitions
        comp_pks = []
        sep1_mult_param, sep2_mult_param = None, None
        if printable:
            sep1_mult_param = max_len_field_name + (TAB_SPACE_COUNT - max_len_field_name % TAB_SPACE_COUNT)
            sep2_mult_param = max_len_field_type + (TAB_SPACE_COUNT - max_len_field_type % TAB_SPACE_COUNT)
        for inx, name in enumerate(field_names):
            field_type = field_types[inx]
            constraints = field_constraints[inx]
            # check for composite primary key
            if count_keys > 1 and constraints and constraints[0] == CONSTRAINT_PRIMARY_KEY:
                comp_pks.append(name)
                constraints = constraints[1:]
            constraints = ' '.join(constraints) if constraints else ''
            line = gen_col_line('', name, ' ', field_type, ' ', constraints, inx, count_fields)
            if printable:
                sep0 = ' ' * TAB_SPACE_COUNT
                sep1 = ' ' * (sep1_mult_param - len(name))
                sep2 = ' ' * (sep2_mult_param - len(field_type))
                printable_line = gen_col_line(sep0, name, sep1, field_type, sep2, constraints, inx, count_fields)
                add_line(line, printable_line)
            else:
                add_line(line)

        # composite primary keys (if more than 1)
        if comp_pks:
            # add comma to previous line
            lines[-1] += ','
            # add PK line
            line = FORMAT_CREATE_TABLE_LINE_PKS % {'sep0': '', 'PKs': ', '.join(comp_pks)}
            if printable:
                printable_lines[-1] += ','
                sep0 = ' ' * TAB_SPACE_COUNT
                printable_line = FORMAT_CREATE_TABLE_LINE_PKS % {'sep0': sep0, 'PKs': ', '.join(comp_pks)}
                add_line(line, printable_line)
            else:
                add_line(line)
        # foreign key constraints (if any)
        if relation_fields:
            for inx, rel_name in enumerate(relation_fields):
                # add comma to previous line
                lines[-1] += ','            
                if printable:
                    printable_lines[-1] += ','
                src_name = relation_sources[inx]
                # add FK line
                line = FORMAT_CREATE_TABLE_LINE_FK % {'sep0': '', 'FK': rel_name, 'src': src_name}
                if printable:
                    sep0 = ' ' * TAB_SPACE_COUNT
                    printable_line = FORMAT_CREATE_TABLE_LINE_FK % {'sep0': sep0, 'FK': rel_name, 'src': src_name}
                    add_line(line, printable_line)
                else:
                    add_line(line)
                rel_constr = relation_constraints[inx]
                if rel_constr:
                    for constr in rel_constr:
                         # add FK line
                        line = FORMAT_CREATE_TABLE_LINE_FK_CONSTR % {'sep0': '', 'constraint': constr}
                        if printable:
                            sep0 = ' ' * (TAB_SPACE_COUNT * 2)
                            printable_line = FORMAT_CREATE_TABLE_LINE_FK_CONSTR % {'sep0': sep0, 'constraint': constr}
                            add_line(line, printable_line)
                        else:
                            add_line(line)               

        # ending lines
        add_line(FORMAT_CREATE_TABLE_LINE_END)
        add_line(FORMAT_CREATE_TABLE_LINE_TRUNC % {'tableName': table_name})

        # generate statement text
        statement = ' '.join(lines)

        if printable:
            if py_mode:
                # handle python mode
                print_statement = gen_py_printable_statement(STATEMENT_TYPE_CREATE_TABLE, table_name, printable_lines, py_prefix)
                if py_func:
                    print_statement += f"\n{gen_py_db_create_table_function(table_name, py_prefix)}"
            else:
                print_statement = gen_printable_statement(printable_lines)

    return statement, message, print_statement


def generate_statement_insert_into(
        table_descriptor=None,
        column_name=None,
        where=None,
        printable=False,
        py_mode=False,
        py_prefix=None,
        py_func=None,
    ):
    
    statement = ''
    print_statement = ''

    # process table descriptor
    success, message, data = parse_table_descriptor_table_name_fields(table_descriptor)

    if success:

        table_name, field_names, count_fields, max_len_field_name = data

        lines = []
        printable_lines = []

        def add_line(line, printable_line=None):
            lines.append(line)
            if printable: printable_lines.append(printable_line if printable_line else line)

        # statement start line
        add_line(FORMAT_INSERT_INTO_LINE_START % {'tableName': table_name})

        # field definitions
        for inx, name in enumerate(field_names):
            line = FORMAT_INSERT_INTO_LINE_COL('', name, '' if inx==(count_fields-1) else ',')
            if printable:
                printable_line = FORMAT_INSERT_INTO_LINE_COL(' ' * TAB_SPACE_COUNT, name, '' if inx==(count_fields-1) else ',')
                add_line(line, printable_line)
            else:
                add_line(line)

        # ending line
        add_line(FORMAT_INSERT_INTO_LINE_END)

        # generate statement text
        statement = ' '.join(lines)

        if printable:
            if py_mode:
                # handle python mode
                print_statement = gen_py_printable_statement(STATEMENT_TYPE_INSERT_INTO, table_name, printable_lines, py_prefix)
                if py_func:
                    print_statement += f"\n{gen_py_db_insert_update_function(table_name, field_names, count_fields, max_len_field_name, py_prefix)}"
            else:
                print_statement = gen_printable_statement(printable_lines)

    return statement, message, print_statement


def generate_statement_merge(
        table_descriptor=None,
        column_name=None,
        where=None,
        printable=False,
        py_mode=False,
        py_prefix=None,
        py_func=None,
    ):
    
    statement = ''
    print_statement = ''

    # process table descriptor
    success, message, data = parse_table_descriptor_all(table_descriptor)

    if success:

        table_name, dependent_tables, field_names, field_types, field_constraints,\
            field_merge_types, relation_fields, relation_sources, relation_constraints,\
                count_fields, count_keys, max_len_field_name, max_len_field_type, count_fks,\
                    key_fields, non_key_fields, merge_on_fields, non_merge_on_fields,\
                        non_merge_on_fields_merge_types = data

        lines = []
        printable_lines = []

        def add_line(line, printable_line=None):
            lines.append(line)
            if printable: printable_lines.append(printable_line if printable_line else line)

        # statement start lines
        add_line(FORMAT_MERGE_LINE_START % {'tableName': proper_table_name(table_name)})
        add_line(FORMAT_MERGE_LINE_USING)

        # all fields
        for inx, name in enumerate(field_names):
            line = FORMAT_MERGE_LINE_COL('', name, '' if inx==(count_fields-1) else ',')
            if printable:
                printable_line = FORMAT_MERGE_LINE_COL(' ' * TAB_SPACE_COUNT, name, '' if inx==(count_fields-1) else ',')
                add_line(line, printable_line)
            else:
                add_line(line)

        # on keys
        add_line(FORMAT_MERGE_LINE_ON)
        count_merge_on_fields = len(merge_on_fields)
        for inx, name in enumerate(merge_on_fields):
            line = FORMAT_MERGE_LINE_ON_KEY % {
                "sep0": '',
                "colName": name,
                "_and": '' if inx==(count_merge_on_fields-1) else ' and'
            }
            if printable:
                printable_line = FORMAT_MERGE_LINE_ON_KEY % {
                    "sep0": ' ' * TAB_SPACE_COUNT,
                    "colName": name,
                    "_and": '' if inx==(count_merge_on_fields-1) else ' and'
                }
                add_line(line, printable_line)
            else:
                add_line(line)
        add_line(FORMAT_MERGE_LINE_ON_END)

        # matched
        count_non_merge_on_fields = len(non_merge_on_fields)
        if count_non_merge_on_fields > 0:
            add_line(FORMAT_MERGE_LINE_MATCH)
        for inx, name in enumerate(non_merge_on_fields):
            merge_type = non_merge_on_fields_merge_types[inx]
            line = (FORMAT_MERGE_LINE_MATCH_ACCUMUL if merge_type==COL_MERGE_TYPE_ACCUMLATE else FORMAT_MERGE_LINE_MATCH_REPLACE) % {
                "sep0": '',
                "colName": name,
                "comma": '' if inx==(count_non_merge_on_fields-1) else ','
            }
            if printable:
                printable_line = (FORMAT_MERGE_LINE_MATCH_ACCUMUL if merge_type==COL_MERGE_TYPE_ACCUMLATE else FORMAT_MERGE_LINE_MATCH_REPLACE) % {
                    "sep0": ' ' * TAB_SPACE_COUNT,
                    "colName": name,
                    "comma": '' if inx==(count_non_merge_on_fields-1) else ','
                }
                add_line(line, printable_line)
            else:
                add_line(line)
        
        # not matched
        add_line(FORMAT_MERGE_LINE_NMATCH)
        for inx, name in enumerate(field_names):
            line = FORMAT_MERGE_LINE_NMATCH_FIELD % {
                "sep0": '',
                "colName": name,
                "comma": '' if inx==(count_fields-1) else ','
            }
            if printable:
                printable_line = FORMAT_MERGE_LINE_NMATCH_FIELD % {
                    "sep0": ' ' * TAB_SPACE_COUNT,
                    "colName": name,
                    "comma": '' if inx==(count_fields-1) else ','
                }
                add_line(line, printable_line)
            else:
                add_line(line)
        add_line(FORMAT_MERGE_LINE_NMATCH_VALUES)
        for inx, name in enumerate(field_names):
            line = FORMAT_MERGE_LINE_NMATCH_VAL('', name, '' if inx==(count_fields-1) else ',')
            if printable:
                printable_line = FORMAT_MERGE_LINE_NMATCH_VAL(' ' * TAB_SPACE_COUNT, name, '' if inx==(count_fields-1) else ',')
                add_line(line, printable_line)
            else:
                add_line(line)


        # ending line
        add_line(FORMAT_MERGE_LINE_END)

        # generate statement text
        statement = ' '.join(lines)

        if printable:
            if py_mode:
                # handle python mode
                print_statement = gen_py_printable_statement(STATEMENT_TYPE_MERGE, table_name, printable_lines, py_prefix)
                if py_func:
                    print_statement += f"\n{gen_py_db_insert_update_function(table_name, field_names, count_fields, max_len_field_name, py_prefix)}"
            else:
                print_statement = gen_printable_statement(printable_lines)

    return statement, message, print_statement


def generate_statement_delete_from(
        table_descriptor=None,
        column_name=None,
        where=None,
        printable=False,
        py_mode=False,
        py_prefix=None,
        py_func=None,
    ):
    
    # where = WHERE statement
    
    success = False
    statement = ''
    print_statement = ''

    if where:

        # process table descriptor
        success, message, table_name = parse_table_descriptor_table_name(table_descriptor)
        
        if success:

            lines = []
            printable_lines = []

            def add_line(line, printable_line=None):
                lines.append(line)
                if printable: printable_lines.append(printable_line if printable_line else line)

            # delete from line
            add_line(FORMAT_DELETE_FROM % {'tableName': table_name})
            
            # where line
            line = FORMAT_DELETE_FROM_WHERE % {
                'sep0': '',
                'where': where
            }
            if printable:
                print_line = FORMAT_DELETE_FROM_WHERE % {
                    'sep0': ' ' * TAB_SPACE_COUNT,
                    'where': where
                } 
                add_line(line, print_line)
            else:
                add_line(line)

            # generate statement text
            statement = ' '.join(lines)

            if printable:
                if py_mode:
                    # handle python mode
                    print_statement = gen_py_printable_statement(STATEMENT_TYPE_DELETE_FROM, table_name, printable_lines, py_prefix)
                    if py_func:
                        print_statement += f"\n{gen_py_db_delete_function(table_name, column_name, py_prefix)}"
                else:
                    print_statement = gen_printable_statement(printable_lines)
    else:
        message = "`where` argument is required for DELETE_FROM statement"
    return statement, message, print_statement


def generate_statement_rename_column(
        table_descriptor=None,
        column_name=None,
        where=None,
        printable=False,
        py_mode=False,
        py_prefix=None,
        py_func=None,
    ):
    
    statement = ''
    print_statement = ''

    # process table descriptor
    success, message, table_name = parse_table_descriptor_table_name(table_descriptor)

    if success:

        lines = []
        printable_lines = []

        def add_line(line, printable_line=None):
            lines.append(line)
            if printable: printable_lines.append(printable_line if printable_line else line)

        # rename line
        add_line(FORMAT_RENAME_COLUMN(table_name))

        # generate statement text
        statement = ' '.join(lines)

        if printable:
            if py_mode:
                # handle python mode
                print_statement = gen_py_printable_statement(STATEMENT_TYPE_RENAME_COLUMN, table_name, printable_lines, py_prefix)
                if py_func:
                    print_statement += f"\n{gen_py_db_rename_column_function(table_name, py_prefix)}"
            else:
                print_statement = gen_printable_statement(printable_lines)

    return statement, message, print_statement


STATEMENT_TYPE_TO_GENERATOR = {
    STATEMENT_TYPE_CREATE_TABLE:    generate_statement_create_table,
    STATEMENT_TYPE_INSERT_INTO:     generate_statement_insert_into,
    STATEMENT_TYPE_MERGE:           generate_statement_merge,
    STATEMENT_TYPE_DELETE_FROM:     generate_statement_delete_from,
    STATEMENT_TYPE_RENAME_COLUMN:   generate_statement_rename_column,
}


def ms_sql_statement(
        st_type,
        table_descriptor,
        column_name=None,
        where=None,
        out_file=None,
        out_overwrite=False,
        printable=False,
        py_mode=False,
        py_prefix=None,
        py_func=False,
    ):

    if st_type not in OPTIONS_STATEMENT_TYPE:
        statement = None
        print_statement = None
        message = f"invalid/unsupported statement type: \'{st_type}\' (options: {', '.join(OPTIONS_STATEMENT_TYPE)})"
    else:
        # call parser of relevant statement
        generator_function = STATEMENT_TYPE_TO_GENERATOR[st_type]
        statement, message, print_statement = generator_function(
            table_descriptor=table_descriptor,
            column_name=column_name,
            where=where,
            printable=printable,
            py_mode=py_mode,
            py_prefix=py_prefix,
            py_func=py_func
        )

    # write result to output file
    if out_file and statement:
        write_statement_to_output_file(
            out_file, 
            print_statement if printable else statement, 
            FILE_WRITE_MODE_OVERWITE if out_overwrite else FILE_WRITE_MODE_APPEND, 
            py_mode
        )
    
    return (statement, message, print_statement) if printable else (statement, message)


def get_table_descriptor_from_file(file):
    # check existence of table descriptor file
    import os
    if not os.path.exists(file):
        return False, f'file \'{file}\' does not exist', None
    td = None
    success = False
    message = ''
    with open(file, 'r') as fd: 
        text = fd.read()
        import json
        try:
            td = json.loads(text)
        except (json.JSONDecodeError, TypeError):
            message = 'file \'{file}\' is not a valid json file'
        if td and type(td)==dict:
            success = True
        else:
            message = 'could not decode json from file \'{file}\''
    return success, message, td


def write_statement_to_output_file(out_file, statement, write_mode=FILE_WRITE_MODE_APPEND, py_mode=False):
    if write_mode == FILE_WRITE_MODE_APPEND:
        # check for content in file
        import os
        if os.path.exists(out_file) and os.path.getsize(out_file) != 0:
            statement = f"\n{'#' if py_mode else '--'} {'-'*90}\n{statement}"
    with open(out_file, write_mode) as out:
        out.write(statement)


"""
Table Descriptor parser functions
"""

from functools import cmp_to_key

try:
    # if running from a parent package (python -m app.py ...) i.e. relative import
    from .options import (
        CONSTRAINT_PRIMARY_KEY,
        DEFAULT_COL_MERGE_TYPE,
        OPTIONS_COL_MERGE_TYPE,
        OPTIONS_COL_RELATION_UPDEL,
        COL_MERGE_TYPE_SKIP
    )
except ImportError:
    # running from app folder (python app.py ...)
    from options import (
        CONSTRAINT_PRIMARY_KEY,
        DEFAULT_COL_MERGE_TYPE,
        OPTIONS_COL_MERGE_TYPE,
        OPTIONS_COL_RELATION_UPDEL,
        COL_MERGE_TYPE_SKIP
    )


def custom_compare_keys(a, b):
    # assuming a, b is either Type int or str
    a = a[0]
    b = b[0]
    try:
        a = int(a)
    except ValueError:
        a_is_int = False
    else:
        a_is_int = True
    try:
        b = int(b)
    except ValueError:
        if a_is_int:
            # a is a number but b is not
            return -1
        # a is a number but b is not
        return man_cmp(a.lower(), b.lower())
    else:
        if a_is_int:
            # a and b both are numbers
            return man_cmp(a, b)
        # a is not a number but b is
        return 1


def man_cmp(a, b):
    return (a > b) - (a < b) 


def parse_table_descriptor_all(td):
    """Returns [success(bool), message(str), parsed_fields(array)]"""
    
    # observations
    count_fields = 0
    count_keys = 0
    max_len_field_name = 0
    max_len_field_type = 0
    count_fks = 0
    # fields to parse
    table_name = None
    dependent_tables = []
    # ordered by key
    field_names = []
    field_types = []
    field_constraints = []
    field_merge_types = []
    # fk entries
    relation_fields = []
    relation_sources = []
    relation_constraints = []
    # merge related
    merge_on_fields = []
    non_merge_on_fields = []
    non_merge_on_fields_merge_types = []
    # key and non-key fields
    key_fields = []
    non_key_fields = []
    # obtain table name (required)
    table_name = td.get('name', None)
    if (not table_name) or (type(table_name) != str):
        return False, f"table name invalid or not provided", []
    table_name = table_name.strip()
    # obtain dependent table names (optional)
    dependent_tables = td.get('dependents', [])
    if dependent_tables:
        for inx, name in enumerate(dependent_tables):
            try:
                dependent_tables[inx] = str(name)
            except (ValueError, TypeError):
                return False, f"bad argument in field 'dependents'", []
    # obtain merge_on fields (required for merge)
    merge_on = td.get('merge_on', [])
    if merge_on:
        for inx, merge_name in enumerate(merge_on):
            try:
                merge_on[inx] = str(merge_name)
            except (ValueError, TypeError):
                return False, f"bad argument in field 'merge_on'", []
    # obtain field names & params (required at least one)
    fields = td.get('fields', [])
    if not fields:
        return False, f"table 'fields' not provided", []
    # sort the fields by key
    fields = sorted(fields.items(), key=cmp_to_key(custom_compare_keys))
    for inx, field in fields:
        # obtain field name (required)
        field_name = field.get('name', None)
        if (not field_name) or (type(field_name) != str):
            return False, f"column name not provided for field with key '{inx}'", []
        field_name = field_name.strip()
        field_names.append(field_name)
        # observe count_fields
        count_fields += 1
        # observe max_len_field_name
        len_field_name = len(field_name)
        if len_field_name > max_len_field_name:
            max_len_field_name = len_field_name
        # obtain field type (required)
        field_type = field.get('type', None)
        if (not field_type) or (type(field_type) != str):
            return False, f"column type not provided for field with key '{inx}' ({field_name})", []
        field_type = field_type.strip().upper()
        # obtain field type args (required)
        type_args = field.get('type_args', None)
        if type_args:
            for inx, arg in enumerate(type_args):
                try:
                    arg = str(arg)
                    type_args[inx] = arg.strip().upper()
                except (ValueError, TypeError):
                    return False, f"bad field type arguments for field with key '{inx}' ({field_name})", []
            field_type += f"({','.join(type_args)})"
        field_type = field_type.strip().upper()
        field_types.append(field_type)
        # observe max_len_field_type
        len_field_type = len(field_type)
        if len_field_type > max_len_field_type:
            max_len_field_type = len_field_type       
        # check for constraints
        constraints = []
        is_key = field.get('key', None)
        if is_key:
            # CONSTRAINT_PRIMARY_KEY (if is_key) will always be the first constraint
            constraints.append(CONSTRAINT_PRIMARY_KEY)
            key_fields.append(field_name)
            # observe count_keys
            count_keys += 1
        else:
            non_key_fields.append(field_name)
        # obtain additional constraints
        constraint_list = field.get('constraints', None)
        if constraint_list:
            for inx, arg in enumerate(constraint_list):
                try:
                    arg = str(arg)
                    constraint_list[inx] = arg.strip().upper()
                except (ValueError, TypeError):
                    return False, f"bad constraint value for field with key '{inx}' ({field_name})", []
            constraints += constraint_list
        field_constraints.append(constraints)
        # obtain field merge_type (optional)
        merge_type = field.get('merge_type', None)
        if not merge_type:
            merge_type = DEFAULT_COL_MERGE_TYPE
        elif type(merge_type) == str:
            merge_type = merge_type.strip().upper()
            if merge_type not in OPTIONS_COL_MERGE_TYPE:
                return False, f"bad merge_type argument for field with key '{inx}' ({field_name})", []
        else:
            return False, f"bad merge_type argument for field with key '{inx}' ({field_name})", []
        field_merge_types.append(merge_type)
        # check if field is in merge_on
        if field_name in merge_on:
            merge_on_fields.append(field_name)
        elif merge_type != COL_MERGE_TYPE_SKIP:
            non_merge_on_fields.append(field_name)
            non_merge_on_fields_merge_types.append(merge_type)
        # obtain relation fields
        relation = field.get('relation', None)
        if relation:
            rel_enforce = relation.get('enforce', True)
            if rel_enforce:
                # observe count_fks
                count_fks += 1
                #  obtain relation table and column
                rel_table = relation.get('table', None)
                if (not rel_table) or (type(rel_table) != str):
                    return False, f"relation table name invalid or not provided\
                        for field with key '{inx}' ({field_name})", []
                rel_table = rel_table.strip()
                rel_col = relation.get('column', None)
                if (not rel_col) or (type(rel_col) != str):
                    return False, f"relation column name invalid or not provided\
                        for field with key '{inx}' ({field_name})", []
                rel_col = rel_col.strip()
                relation_source = f"{rel_table}({rel_col})"
                relation_sources.append(relation_source)
                relation_fields.append(field_name)
                # obtain relation constraints
                rel_constraints = []
                rel_on_delete = relation.get('on_delete', None)
                if rel_on_delete:
                    if type(rel_on_delete) == str:
                        rel_on_delete = rel_on_delete.strip().upper()
                        if rel_on_delete not in OPTIONS_COL_RELATION_UPDEL:
                            return False, f"invalide value of relation on_delete constraint \
                                for field with key '{inx}' ({field_name})", []
                        else:
                            rel_constraints.append(f"ON DELETE {rel_on_delete}")
                    else:
                        return False, f"invalid value of relation on_delete constraint \
                            for field with key '{inx}' ({field_name})", []
                rel_on_update = relation.get('on_update', None)
                if rel_on_update:
                    if type(rel_on_update) == str:
                        rel_on_update = rel_on_update.strip().upper()
                        if rel_on_update not in OPTIONS_COL_RELATION_UPDEL:
                            return False, f"invalide value of relation on_update constraint \
                                for field with key '{inx}' ({field_name})", []
                        else:
                            rel_constraints.append(f"ON UPDATE {rel_on_update}")
                    else:
                        return False, f"invalid value of relation on_update constraint \
                            for field with key '{inx}' ({field_name})", []
                relation_constraints.append(rel_constraints)

    return True, 'parsed successfully', [
        table_name, dependent_tables, field_names, field_types, field_constraints, 
        field_merge_types, relation_fields, relation_sources, relation_constraints,
        count_fields, count_keys, max_len_field_name, max_len_field_type, count_fks,
        key_fields, non_key_fields, merge_on_fields, non_merge_on_fields,
        non_merge_on_fields_merge_types
    ]


def parse_table_descriptor_table_name(td):
    """Returns [success(bool), message(str), table_name(str)]"""
    
    table_name = td.get('name', None)
    if (not table_name) or (type(table_name) != str):
        return False, f"table name invalid or not provided", []
    table_name = table_name.strip()
    
    return True, 'parsed successfully', table_name


def parse_table_descriptor_table_name_fields(td):
    """Returns [success(bool), message(str), parsed_fields(array)]"""
    
    # observations
    count_fields = 0
    max_len_field_name = 0
    # fields to parse
    table_name = None
    # ordered by key
    field_names = []
    # obtain table name (required)
    table_name = td.get('name', None)
    if (not table_name) or (type(table_name) != str):
        return False, f"table name invalid or not provided", []
    table_name = table_name.strip()
    # obtain field names & params (required at least one)
    fields = td.get('fields', None)
    if not fields:
        return False, f"table 'fields' not provided", []
    # sort the fields by key
    fields = sorted(fields.items(), key=cmp_to_key(custom_compare_keys))
    for inx, field in fields:
        # obtain field name (required)
        field_name = field.get('name', None)
        if (not field_name) or (type(field_name) != str):
            return False, f"column name not provided for field with key '{inx}'", []
        field_name = field_name.strip()
        field_names.append(field_name)
        # observe count_fields
        count_fields += 1
        # observe max_len_field_name
        len_field_name = len(field_name)
        if len_field_name > max_len_field_name:
            max_len_field_name = len_field_name

    return True, 'parsed successfully', [
        table_name, field_names, count_fields, max_len_field_name,
    ]



"""
Logging-related sruff.
"""

import os
from pathlib import Path
import logging


LOG_FILE = 'logs/sql_generator.log'
# DEFAULT: don't write to terminal
PRINT_TERMINAL = False
# DEFAULT: always write to log
SKIP_LOG_WRITE = False


logger = logging.getLogger(__name__)

def init_logger(log_file=LOG_FILE):
    # generate log file path if does not exist
    log_file_directory = Path(log_file).parent.resolve()
    if log_file_directory and not os.path.exists(log_file_directory):
        os.makedirs(log_file_directory)
    # set log level
    logger.setLevel(logging.INFO)
    # log format
    formatter = logging.Formatter('%(asctime)s [%(levelname)s]  %(message)s', '%Y-%m-%d %H:%M:%S')    
    # log file handler
    filehandler = logging.FileHandler(log_file, 'a')
    filehandler.setFormatter(formatter)
    # remove previous log file handlers
    for handler in logger.handlers[:]:
        if isinstance(handler, logging.FileHandler):
            logger.removeHandler(handler)
    # set log file handler
    logger.addHandler(filehandler)

init_logger()

def sqllogger(*string):
    if PRINT_TERMINAL:
        print(*string) 
    if not SKIP_LOG_WRITE:
        logger.info(*string)

def print_to_terminal_ON():
    global PRINT_TERMINAL
    PRINT_TERMINAL = True

def print_to_terminal_OFF():
    global PRINT_TERMINAL
    PRINT_TERMINAL = False

def write_to_log_ON():
    global SKIP_LOG_WRITE
    SKIP_LOG_WRITE = False

def write_to_log_OFF():
    global SKIP_LOG_WRITE
    SKIP_LOG_WRITE = True



"""
MS SQL Query Syntax
"""

try:
    # if running from a parent package (python -m app.py ...) i.e. relative import
    from .options import (
        STATEMENT_TYPE_ALT_NAMES, STATEMENT_TYPE_CREATE_TABLE, STATEMENT_TYPE_DELETE_FROM, 
        STATEMENT_TYPE_INSERT_INTO, STATEMENT_TYPE_MERGE, STATEMENT_TYPE_RENAME_COLUMN
    )
except ImportError:
    # running from app folder (python app.py ...)
    from options import (
        STATEMENT_TYPE_ALT_NAMES, STATEMENT_TYPE_CREATE_TABLE, STATEMENT_TYPE_DELETE_FROM, 
        STATEMENT_TYPE_INSERT_INTO, STATEMENT_TYPE_MERGE, STATEMENT_TYPE_RENAME_COLUMN
    )


MS_SQL_TABLE_PREFIX = "dbo."
TAB_SPACE_COUNT = 4


def proper_table_name(name):
    return f"{MS_SQL_TABLE_PREFIX}{name}"


# MS SQL FORMATS

FORMAT_DROP_TABLE_IF_EXISTS = "IF OBJECT_ID('%(tableName)s', 'U') IS NOT NULL DROP TABLE %(tableName)s;"

FORMAT_CREATE_TABLE_LINE_START = "CREATE TABLE %(tableName)s ("
FORMAT_CREATE_TABLE_LINE_COL = "%(sep0)s%(colName)s%(sep1)s%(colType)s%(sep2)s%(constraints)s%(comma)s"
FORMAT_CREATE_TABLE_LINE_PKS = "%(sep0)sPRIMARY KEY (%(PKs)s)"
FORMAT_CREATE_TABLE_LINE_FK = "%(sep0)sFOREIGN KEY (%(FK)s) REFERENCES %(src)s"
FORMAT_CREATE_TABLE_LINE_FK_CONSTR = "%(sep0)s%(constraint)s"
FORMAT_CREATE_TABLE_LINE_END = ");"
FORMAT_CREATE_TABLE_LINE_TRUNC = "TRUNCATE TABLE %(tableName)s;"

FORMAT_INSERT_INTO_LINE_START = "INSERT INTO %(tableName)s VALUES ("
def FORMAT_INSERT_INTO_LINE_COL(sep0, colName, comma): return f"{sep0}%({colName})s{comma}"
FORMAT_INSERT_INTO_LINE_END = ");"

FORMAT_MERGE_LINE_START = "MERGE %(tableName)s AS tgt"
FORMAT_MERGE_LINE_USING = "USING (SELECT"
def FORMAT_MERGE_LINE_COL(sep0, colName, comma): return f"{sep0}%({colName})s {colName}{comma}"
FORMAT_MERGE_LINE_ON = ") AS src ON ("
FORMAT_MERGE_LINE_ON_KEY = "%(sep0)ssrc.%(colName)s = tgt.%(colName)s%(_and)s"
FORMAT_MERGE_LINE_ON_END = ")"
FORMAT_MERGE_LINE_MATCH = "WHEN MATCHED THEN UPDATE SET"
FORMAT_MERGE_LINE_MATCH_REPLACE = "%(sep0)stgt.%(colName)s = src.%(colName)s%(comma)s"
FORMAT_MERGE_LINE_MATCH_ACCUMUL = "%(sep0)stgt.%(colName)s = tgt.%(colName)s+src.%(colName)s%(comma)s"
FORMAT_MERGE_LINE_NMATCH = "WHEN NOT MATCHED THEN INSERT ("
FORMAT_MERGE_LINE_NMATCH_FIELD = "%(sep0)s%(colName)s%(comma)s"
FORMAT_MERGE_LINE_NMATCH_VALUES = ") VALUES ("
def FORMAT_MERGE_LINE_NMATCH_VAL(sep0, colName, comma): return f"{sep0}%({colName})s{comma}"
FORMAT_MERGE_LINE_END = ");"

FORMAT_DELETE_FROM = "DELETE FROM %(tableName)s"
FORMAT_DELETE_FROM_WHERE = "%(sep0)sWHERE %(where)s;"

def FORMAT_RENAME_COLUMN(table): return f"EXEC sp_rename '{table}.%(column)s', %(newName)s;"


# PYTHON FUNCTION FORMATS

FORMAT_PY_FUNC_CREATE_TABLE = f"# {'-'*25}delete{'-'*25}\ncursor, conn = None, None\n" + f"# {'-'*(25*2+len('delete'))}\n" + \
    "def create_table_%(tableLower)s(commit=True):\n" + \
        f"{' '*TAB_SPACE_COUNT}cursor.execute(" + "%(prefix_)s" + \
            STATEMENT_TYPE_ALT_NAMES[STATEMENT_TYPE_CREATE_TABLE]+"_%(tableUpper)s)\n" + \
            f"{' '*TAB_SPACE_COUNT}if commit:\n{' '*(TAB_SPACE_COUNT*2)}conn.commit()"

FORMAT_PY_FUNC_INSERT_UPDATE = f"# {'-'*25}delete{'-'*25}\ncursor, conn = None, None\n" + "%(prefix_)s" + \
    STATEMENT_TYPE_ALT_NAMES[STATEMENT_TYPE_INSERT_INTO] +"_%(tableUpper)s = None\n%(prefix_)s" + \
        STATEMENT_TYPE_ALT_NAMES[STATEMENT_TYPE_MERGE] + "_%(tableUpper)s = None\n" + \
            f"# {'-'*(25*2+len('delete'))}\n" + \
                "def insert_into_%(tableLower)s(values, update_if_exists=False, commit=True):\n" + \
                    f"{' '*TAB_SPACE_COUNT}query = (\n{' '*(TAB_SPACE_COUNT*2)}" + \
                        "%(prefix_)s"+STATEMENT_TYPE_ALT_NAMES[STATEMENT_TYPE_MERGE] + \
                            "_%(tableUpper)s if update_if_exists else %(prefix_)s" + \
                                STATEMENT_TYPE_ALT_NAMES[STATEMENT_TYPE_INSERT_INTO]+"_%(tableUpper)s" + \
                                f"\n{' '*TAB_SPACE_COUNT}) +>?X$;~;^&$ " + "{\n" + "%(fieldLines)s" + f"{' '*TAB_SPACE_COUNT}" + "}\n" + \
                                    f"{' '*TAB_SPACE_COUNT}cursor.execute(query)\n" + \
                                        f"{' '*TAB_SPACE_COUNT}if commit:\n{' '*(TAB_SPACE_COUNT*2)}conn.commit()"
FORMAT_PY_FUNC_INSERT_UPDATE_FIELD = f"{' '*(TAB_SPACE_COUNT*2)}" + "\"%(colName)s\":%(sep0)svalues[%(inx)s]%(comma)s\n"

FORMAT_PY_FUNC_DELETE_FROM = f"# {'-'*25}delete{'-'*25}\ncursor, conn = None, None\n" + f"# {'-'*(25*2+len('delete'))}\n" + \
    "def delete_from_%(tableLower)s(id, commit=True):\n" + \
        f"{' '*TAB_SPACE_COUNT}query = "+"%(prefix_)s"+STATEMENT_TYPE_ALT_NAMES[STATEMENT_TYPE_DELETE_FROM] + \
            "_%(tableUpper)s +>?X$;~;^&$ "+"{\"%(column)s\": id}\n" + \
                f"{' '*TAB_SPACE_COUNT}cursor.execute(query)\n" + \
                    f"{' '*TAB_SPACE_COUNT}if commit:\n{' '*(TAB_SPACE_COUNT*2)}conn.commit()"

FORMAT_PY_FUNC_RENAME_COLUMN = f"# {'-'*25}delete{'-'*25}\ncursor, conn = None, None\n" + f"# {'-'*(25*2+len('delete'))}\n" + \
    "def rename_column_%(tableLower)s(column, new_name, commit=True):\n" + \
        f"{' '*TAB_SPACE_COUNT}query = "+"%(prefix_)s"+STATEMENT_TYPE_ALT_NAMES[STATEMENT_TYPE_RENAME_COLUMN] + \
            "_%(tableUpper)s +>?X$;~;^&$ "+"{\n" + \
                f"{' '*(TAB_SPACE_COUNT*2)}"+"\"column\": column,\n"+f"{' '*(TAB_SPACE_COUNT*2)}"+"\"newName\": new_name\n" + \
                f"{' '*TAB_SPACE_COUNT}" + "}\n" + f"{' '*TAB_SPACE_COUNT}cursor.execute(query)\n" + \
                    f"{' '*TAB_SPACE_COUNT}if commit:\n{' '*(TAB_SPACE_COUNT*2)}conn.commit()"



import os
from pathlib import Path
from setuptools import setup
from pip._internal.req import parse_requirements


this_directory = Path(__file__).parent
package_directory = 'mssqlgenerator'
description_file = (this_directory / package_directory / "README.md")
requirements_file = (this_directory / package_directory / "requirements.txt")

long_description = None
if os.path.exists(description_file):
    # read the contents of README file
    long_description = description_file.read_text()

install_requires = None
if os.path.exists(requirements_file):
    # read the contents of requirements file
    reqs = parse_requirements(requirements_file, session="test")
    install_requires = [str(ir.req) for ir in reqs]

# call setup to generate distributable
setup(
    name="mssqlgenerator",
    version="1.1.1-1",
    author="Kamil Javed",
    author_email="kamiljaved98@gmail.com",
    url="https://github.com/kamiljaved/mssqlgenerator",
    packages=["mssqlgenerator"],
    description="Simple T-SQL query generator based on JSON parameters.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.8.10',
    install_requires=install_requires,
)



"""
Variables defining options for:
- statement types
- table descriptor dictionary attribute selectable values
- other stuff
"""


# conf options

FILE_WRITE_MODE_APPEND = 'a+'
FILE_WRITE_MODE_OVERWITE = 'w'

# generator options

STATEMENT_TYPE_CREATE_TABLE     = 'CREATE_TABLE'
STATEMENT_TYPE_INSERT_INTO      = 'INSERT_INTO'
STATEMENT_TYPE_MERGE            = 'MERGE'
STATEMENT_TYPE_DELETE_FROM      = 'DELETE_FROM'
STATEMENT_TYPE_RENAME_COLUMN    = 'RENAME_COLUMN'

OPTIONS_STATEMENT_TYPE = [
    STATEMENT_TYPE_CREATE_TABLE,
    STATEMENT_TYPE_INSERT_INTO,
    STATEMENT_TYPE_MERGE,
    STATEMENT_TYPE_DELETE_FROM,
    STATEMENT_TYPE_RENAME_COLUMN,
]

OPTIONS_STATEMENT_TYPE_LOWER = [
    STATEMENT_TYPE_CREATE_TABLE.lower(),
    STATEMENT_TYPE_INSERT_INTO.lower(),
    STATEMENT_TYPE_MERGE.lower(),
    STATEMENT_TYPE_DELETE_FROM.lower(),
    STATEMENT_TYPE_RENAME_COLUMN.lower(),
]

STATEMENT_TYPE_ALT_NAMES = {
    STATEMENT_TYPE_CREATE_TABLE:    "CREATE_TABLE",
    STATEMENT_TYPE_INSERT_INTO:     "INSERT_VALUES",
    STATEMENT_TYPE_MERGE:           "INSERT_UPDATE_VALUES",
    STATEMENT_TYPE_DELETE_FROM:     "DELETE_FROM",
    STATEMENT_TYPE_RENAME_COLUMN:   "RENAME_COLUMN",
}

COL_MERGE_TYPE_REPLACE          = 'REPLACE'
COL_MERGE_TYPE_ACCUMLATE        = 'ACCUMULATE'
COL_MERGE_TYPE_SKIP             = 'SKIP'

OPTIONS_COL_MERGE_TYPE = [
    COL_MERGE_TYPE_REPLACE,
    COL_MERGE_TYPE_ACCUMLATE,
    COL_MERGE_TYPE_SKIP,
]
DEFAULT_COL_MERGE_TYPE = COL_MERGE_TYPE_REPLACE

COL_RELATION_UPDEL_NO_ACTION    = 'NO ACTION'
COL_RELATION_UPDEL_CASCADE      = 'CASCADE'
COL_RELATION_UPDEL_SET_NULL     = 'SET NULL'
COL_RELATION_UPDEL_SET_DEFAULT  = 'SET DEFAULT'

OPTIONS_COL_RELATION_UPDEL = [
    COL_RELATION_UPDEL_NO_ACTION,
    COL_RELATION_UPDEL_CASCADE,
    COL_RELATION_UPDEL_SET_NULL,
    COL_RELATION_UPDEL_SET_DEFAULT,
]

CONSTRAINT_PRIMARY_KEY = 'PRIMARY KEY'
